!-------------------------------------------------------------------------------

!VERS

! This file is part of Code_Saturne, a general-purpose CFD tool.
!
! Copyright (C) 1998-2021 EDF S.A.
!
! This program is free software; you can redistribute it and/or modify it under
! the terms of the GNU General Public License as published by the Free Software
! Foundation; either version 2 of the License, or (at your option) any later
! version.
!
! This program is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
! FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
! details.
!
! You should have received a copy of the GNU General Public License along with
! this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
! Street, Fifth Floor, Boston, MA 02110-1301, USA.

!-------------------------------------------------------------------------------

!============================================!
!== Pinzo@250902024: Fortran module for ANN prediction ==
module ANNtorch
    use, intrinsic :: iso_c_binding

    implicit none
    private
    interface

    ! Pinzo@25092024: Subroutine for ANN predict
    subroutine ANNpredict(output, input_0, input_1, input_2, input_3, input_4) &
                          bind(C, name="ANNpredict")
        use, intrinsic :: iso_c_binding
        implicit none

        type (c_ptr), value :: output   ! equivalent to pointer C++
        type (c_ptr), value :: input_0
        type (c_ptr), value :: input_1
        type (c_ptr), value :: input_2
        type (c_ptr), value :: input_3
        type (c_ptr), value :: input_4

    end subroutine ANNpredict

    ! Pinzo@26092024: Subroutine for time step redefinition
    subroutine cs_time_step_redefine_cur(nt_cur, t_cur) bind(C, name="cs_time_step_redefine_cur")
        use, intrinsic :: iso_c_binding
        implicit none

        integer(kind=c_int), value :: nt_cur
        real(kind=c_double), value :: t_cur

    end subroutine cs_time_step_redefine_cur
    
    end interface
    public :: ANNpredict, cs_time_step_redefine_cur

end module ANNtorch
!== (END)Pinzo@250902024: Fortran module for ANN prediction ==
!============================================!

!============================================!
!== Pinzo@11092024: Camille's code ==
subroutine Zv_extra_op()
    use paramx
    use pointe
    use numvar
    use optcal
    use cstphy
    use cstnum
    use entsor
    use parall
    use period
    use ppppar
    use ppthch
    use ppincl
    use field
    use mesh
    use lagran
    use cs_c_bindings
    use coincl
    use radiat
    use dimens, only: ndimfb
    use field_operator

    implicit none
    integer iprev
    integer inc , iel

    double precision, allocatable, dimension(:,:) :: grad
    double precision, dimension(:), pointer :: cvar_fp2m
    double precision :: grad_mag, grad_sqr, fp2m

    allocate(grad(3, ncelet))
    call field_get_val_s(ivarfl(isca(ifp2m)), cvar_fp2m)   ! get field values of the scalar cvar_f2pm; isca(ifp2m): index of scalar mixfrac variance; ivarfl(): field id

    ! calculate mixture fraction variance
    iprev = 0        ! 1 = value at previous iteration, 0 = otherwise
    inc = 1          ! 1 = do not increment, 0 = otherwise

    call field_gradient_scalar(ivarfl(isca(ifp2m)), iprev, inc,            &
                               grad)
    
    ! Zv calculation
    do iel = 1, ncel
        grad_mag = sqrt(grad(1,iel)*grad(1,iel)+                           &
                        grad(2,iel)*grad(2,iel)+                           & 
                        grad(3,iel)*grad(3,iel))   ! variance magnitude
        
        grad_sqr = grad_mag*grad_mag

        ! variance Z calculation 
        fp2m = 0.18*(volume(iel)**((2.d0)/(3.d0)))*grad_sqr

        ! secure value by clipping it between 0 and Z(1-Z)
        if (fp2m.lt.0.0) then
            cvar_fp2m(iel) = 0.d0
        else
            cvar_fp2m(iel) = fp2m
        endif
    enddo

    deallocate(grad)   ! release memory

end subroutine Zv_extra_op

!== (END)Pinzo@11092024: Camille's code ==
!============================================!

!============================================!
!== Pinzo@25092025: Camille's code ==
!== ANNupload: Get transported scalar and predict the next time step by ANN ==
subroutine ANN_upload()
    use paramx
    use pointe
    use numvar
    use optcal
    use cstphy
    use cstnum
    use entsor
    use parall
    use period
    use ppppar
    use ppthch
    use ppincl
    use field
    use mesh
    use lagran
    use cs_c_bindings
    use coincl
    use radiat
    use dimens, only: ndimfb
    use field_operator
    use ANNtorch

    ! To prevent Fortran automatic type assignment
    implicit none

    ! Local variable
    double precision, dimension(:), pointer :: activeANN
    integer iel

    real (c_float), target :: output(8)
    real (c_float), target :: input_0(8)
    real (c_float), target :: input_1(8)
    real (c_float), target :: input_2(8)
    real (c_float), target :: input_3(8)
    real (c_float), target :: input_4(8)

    ! Scalar fields plus velocity for ANN input
    double precision, dimension(:), pointer :: rhoANN_0, rhoANN_1, rhoANN_2, rhoANN_3, rhoANN_4
    double precision, dimension(:), pointer :: tempANN_0, tempANN_1, tempANN_2, tempANN_3, tempANN_4
    double precision, dimension(:,:), pointer :: velANN_0, velANN_1, velANN_2, velANN_3, velANN_4
    double precision, dimension(:,:), pointer :: yiANN_0, yiANN_1, yiANN_2, yiANN_3, yiANN_4

    ! Setting up output to the function
    double precision, dimension(:), pointer :: rhoANN_o, tempANN_o
    double precision, dimension(:,:), pointer :: yiANN_o

    ! Calling fields for ANN input
    ! Density
    call field_get_val_s_by_name("rhoANN_0", rhoANN_0)
    call field_get_val_s_by_name("rhoANN_1", rhoANN_1)
    call field_get_val_s_by_name("rhoANN_2", rhoANN_2)
    call field_get_val_s_by_name("rhoANN_3", rhoANN_3)
    call field_get_val_s_by_name("rhoANN_4", rhoANN_4)
    ! Temperature
    call field_get_val_s_by_name("tempANN_0", tempANN_0)
    call field_get_val_s_by_name("tempANN_1", tempANN_1)
    call field_get_val_s_by_name("tempANN_2", tempANN_2)
    call field_get_val_s_by_name("tempANN_3", tempANN_3)
    call field_get_val_s_by_name("tempANN_4", tempANN_4)
    ! Velocity
    call field_get_val_v_by_name("velANN_0", velANN_0)
    call field_get_val_v_by_name("velANN_1", velANN_1)
    call field_get_val_v_by_name("velANN_2", velANN_2)
    call field_get_val_v_by_name("velANN_3", velANN_3)
    call field_get_val_v_by_name("velANN_4", velANN_4)
    ! Species
    call field_get_val_v_by_name("yiANN_0", yiANN_0)
    call field_get_val_v_by_name("yiANN_1", yiANN_1)
    call field_get_val_v_by_name("yiANN_2", yiANN_2)
    call field_get_val_v_by_name("yiANN_3", yiANN_3)
    call field_get_val_v_by_name("yiANN_4", yiANN_4)

    ! Calling key variable for activation of ANN
    call field_get_val_s_by_name("activeANN", activeANN)
    
    do iel = 1, ncel
        if (activeANN(0).ge.1) then
            ! Setting scalars for ANNpredict function
            !t0
            input_0(1) = rhoANN_0(iel)
            input_0(2) = velANN_0(iel,1)
            input_0(3) = velANN_0(iel,2)
            input_0(4) = velANN_0(iel,3)
            input_0(5) = tempANN_0(iel)
            input_0(6) = yiANN_0(iel,1)
            input_0(7) = yiANN_0(iel,2)
            input_0(8) = yiANN_0(iel,3)
            !t1
            input_1(1) = rhoANN_1(iel)
            input_1(2) = velANN_1(iel,1)
            input_1(3) = velANN_1(iel,2)
            input_1(4) = velANN_1(iel,3)
            input_1(5) = tempANN_1(iel)
            input_1(6) = yiANN_1(iel,1)
            input_1(7) = yiANN_1(iel,2)
            input_1(8) = yiANN_1(iel,3)
            !t2
            input_2(1) = rhoANN_2(iel)
            input_2(2) = velANN_2(iel,1)
            input_2(3) = velANN_2(iel,2)
            input_2(4) = velANN_2(iel,3)
            input_2(5) = tempANN_2(iel)
            input_2(6) = yiANN_2(iel,1)
            input_2(7) = yiANN_2(iel,2)
            input_2(8) = yiANN_2(iel,3)
            !t3
            input_3(1) = rhoANN_3(iel)
            input_3(2) = velANN_3(iel,1)
            input_3(3) = velANN_3(iel,2)
            input_3(4) = velANN_3(iel,3)
            input_3(5) = tempANN_3(iel)
            input_3(6) = yiANN_3(iel,1)
            input_3(7) = yiANN_3(iel,2)
            input_3(8) = yiANN_3(iel,3)
            !t4
            input_4(1) = rhoANN_4(iel)
            input_4(2) = velANN_4(iel,1)
            input_4(3) = velANN_4(iel,2)
            input_4(4) = velANN_4(iel,3)
            input_4(5) = tempANN_4(iel)
            input_4(6) = yiANN_4(iel,1)
            input_4(7) = yiANN_4(iel,2)
            input_4(8) = yiANN_4(iel,3)
            
            !call to ANNpredict
            call ANNpredict(c_loc(output), c_loc(input_0), c_loc(input_1), c_loc(input_2),   &
                            c_loc(input_3), c_loc(input_4))
            
            ! Moving ANN output to dedicated arrays
            call field_get_val_s_by_name("rhoANN_o", rhoANN_o)
            call field_get_val_s_by_name("tempANN_o", tempANN_o)
            call field_get_val_v_by_name("yiANN_o", yiANN_o)

            rhoANN_o(iel) = output(1)
            tempANN_o(iel) = output(5)
            yiANN_o(iel,1) = output(6)
            yiANN_o(iel,2) = output(7)
            yiANN_o(iel,3) = output(8)
        endif 
    enddo

    return

end subroutine ANN_upload

!============================================!
!== Pinzo@26092024: Camille's code ==

subroutine cs_f_user_extra_operations &
    ( nvar   , nscal  ,                                              &
    dt     )
        use paramx
        use pointe
        use numvar
        use optcal
        use cstphy
        use cstnum
        use entsor
        use parall
        use period
        use ppppar
        use ppthch
        use ppincl
        use field
        use mesh
        use lagran
        use cs_c_bindings
        use coincl
        use radiat
        use dimens, only: ndimfb
        use field_operator
        use ANNtorch

        implicit none
    
        integer          nvar   , nscal
        double precision dt(ncelet)

        logical ANNactivated

        ! Pinzo@26092024: Initialize variable for ANN activation
        double precision, dimension(:), pointer :: activeANN
        call field_get_val_s_by_name("activeANN", activeANN)

        if (activeANN(0).ge.1) then
            ANNactivated = .TRUE.
        endif
        if (ANNactivated) then
            call ANN_upload()
        endif

        return

    end subroutine cs_f_user_extra_operations
    
!===============================================================================
!> \brief Modify turbulent viscosity
!>
!> This subroutine is called at beginning of each time step
!> after the computation of the turbulent viscosity
!> (physical quantities have already been computed in \ref usphyv).
!>
!> Turbulent viscosity \f$ \mu_T \f$ (kg/(m s)) can be modified.
!>
!> A modification of the turbulent viscosity can lead to very
!> significant differences between solutions and even give wrong
!> results.
!>
!> This subroutine is therefore reserved to expert users.
!
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Arguments
!______________________________________________________________________________.
!  mode           name          role                                           !
!______________________________________________________________________________!
!> \param[in]     nvar          total number of variables
!> \param[in]     nscal         total number of scalars
!> \param[in]     ncepdp        number of cells with head loss
!> \param[in]     ncesmp        number of cells with mass source term
!> \param[in]     icepdc        head loss cell numbering
!> \param[in]     icetsm        numbering of cells with mass source term
!> \param[in]     itypsm        kind of mass source for each variable
!> \param[in]     dt            time step (per cell)
!> \param[in]     ckupdc        work array for head loss terms
!> \param[in]     smacel        values of variables related to mass source
!>                              term. If ivar=ipr, smacel=mass flux
!_______________________________________________________________________________

subroutine usvist &
    ( nvar   , nscal  , ncepdp , ncesmp ,                            &
      icepdc , icetsm , itypsm ,                                     &
      dt     ,                                                       &
      ckupdc , smacel )
   
   !===============================================================================
   
   !===============================================================================
   ! Module files
   !===============================================================================
   
   use paramx
   use numvar
   use optcal
   use cstphy
   use entsor
   use parall
   use period
   use mesh
   use field
   use field_operator
   
   !===============================================================================
   
   implicit none
   
   ! Arguments
   
   integer          nvar   , nscal
   integer          ncepdp , ncesmp
   
   integer          icepdc(ncepdp)
   integer          icetsm(ncesmp), itypsm(ncesmp,nvar)
   
   double precision dt(ncelet)
   double precision ckupdc(6,ncepdp), smacel(ncesmp,nvar)
   
   !===============================================================================
   
   !--------
   ! Formats
   !--------
   
   !----
   ! End
   !----
   
   return
   end subroutine usvist
   